package name.lamperi.orbital


// Latitude = South - North (-90...90)
// Longitude = East West (-180...180)

val EARTH_RADIUS = 6371.0; // KM


data class Point3D (
        val x : Double, val y : Double, val z : Double)

data class Vec3 (
        val x : Double, val y : Double, val z : Double
) {
    fun len() : Double {
        return Math.sqrt(x * x + y * y + z * z)
    }

    fun unit() : Vec3 {
        val l = len()
        return Vec3(x/l, y/l, z/l)
    }

    fun dot(o: Vec3) : Double {
        return x * o.x + y * o.y + z * o.z
    }

    operator fun plus(o : Vec3) : Vec3 {
        return Vec3(x + o.x, y + o.y, z + o.z)
    }

    operator fun minus(o : Vec3) : Vec3 {
        return Vec3(x - o.x, y - o.y, z - o.z)
    }

    operator fun times(d : Double) : Vec3 {
        return Vec3(x * d, y * d, z * d)
    }
}

fun createVec(start : Point3D, end : Point3D) : Vec3 {
    return Vec3(end.x - start.x, end.y - start.y, end.z - start.z)
}

data class SphericalCoordinate (
        val radius : Double, val theta : Double, val phi : Double)

fun toVec(point : Point3D) : Vec3 {
    return Vec3(point.x, point.y, point.z)
}

// Coordinate system
// x = + when lon = 0°, - when lon = 180°
// y = + when lon = 90°, - when lon = -90°
// z = + when lat = 90°, - when lat = -90°

fun rad(degrees : Double) = degrees * Math.PI / 180.0

// let theta = rad(lon) in (-pi, pi)
// let phi = rad(-lat + 90°) in (0, pi)
fun toSpherical(lat : Double, lon : Double, radius : Double = EARTH_RADIUS)
        =  SphericalCoordinate(radius = radius, theta = rad(lat), phi = rad(lon))

// Cartesian coordinates can be calculated using
// x = r * sin(theta) * cos(phi)
// y = r * sin(theta) * sin(phi)
// z = r * cos(theta)
fun toCartesian(c : SphericalCoordinate)
        = Point3D(
        x = c.radius * Math.cos(c.theta) * Math.cos(c.phi),
        y = c.radius * Math.cos(c.theta) * Math.sin(c.phi),
        z = c.radius * Math.sin(c.theta))

fun toCartesian(lat : Double, lon : Double)
        = toCartesian(toSpherical(lat, lon))

fun toCartesian(lat : Double, lon : Double, height : Double) : Point3D
        = toCartesian(toSpherical(lat, lon, EARTH_RADIUS + height))


// https://en.wikipedia.org/wiki/Line–sphere_intersection
// Solutions are of quadratic form
// l = unit vector from point 0 to point 1
// o = point 0
// c = [0,0,0]
// r = EARTH_RADIUS
// Calculate (l · o) +- Math.sqrt((l · o)^2 - (o · o)  + r^2)
// If it is non-zero, then the route cannot be made
fun intersectsEarth(p0 : Point3D, p1 : Point3D) : Boolean {
    val dist = createVec(p0, p1).len()
    val l = createVec(p0, p1).unit()
    val o = toVec(p0)

    val l_dot_o = l.dot(o)
    val o_dot_o = o.dot(o)
    val dis = l_dot_o * l_dot_o - o_dot_o + EARTH_RADIUS * EARTH_RADIUS
    val d1 = - l_dot_o + Math.sqrt(dis)
    val d2 = - l_dot_o - Math.sqrt(dis)

    val sol1 = o + l * d1
    val sol2 = o + l * d2

    val sigma = 0
    val sol1intersects = d1 > sigma && dist > d1 - sigma
    val sol2intersects = d2 > sigma && dist > d2 - sigma

    //println("$p0 $p1 $dist => $d1 $sol1intersects $sol1 and $d2 $sol2intersects $sol2")

    return sol1intersects || sol2intersects
}

// SAT0 Point3D(x=3372.4860600513102, y=5850.5619038216755, z=-1254.2621619796448)
// SAT5 Point3D(x=-60.473722764130805, y=408.66870956674063, z=-6972.741992625009)
// Start Point3D(x=-70.20728008240212, y=132.20869520700555, z=-6369.241147792654)
// End Point3D(x=215.17891267006544, y=-3374.1448308119298, z=5399.859784869165))

fun main(args : Array<String>) {
    val SAT0 = Point3D(x=3372.4860600513102, y=5850.5619038216755, z=-1254.2621619796448)
    val SAT5 = Point3D(x=-60.473722764130805, y=408.66870956674063, z=-6972.741992625009)
    val Start = Point3D(x=-70.20728008240212, y=132.20869520700555, z=-6369.241147792654)
    val End = Point3D(x=215.17891267006544, y=-3374.1448308119298, z=5399.859784869165)

    println(intersectsEarth(Start, SAT5))
    println(intersectsEarth(SAT5, Start))

    println(intersectsEarth(End, SAT5))
    println(intersectsEarth(SAT5, End))

    println(intersectsEarth(Start, SAT0))
    println(intersectsEarth(SAT0, Start))

    println(intersectsEarth(End, SAT0))
    println(intersectsEarth(SAT0, End))

    println(intersectsEarth(Start, End))
    println(intersectsEarth(End, Start))
}