package name.lamperi.orbital

import java.util.LinkedList

data class Graph (
        val nodes : Map<String, Node>)

data class Node (
        val info : NamedPoint3D, val neighbors : Set<String>)

data class NamedPoint3D (
        val name : String, val point : Point3D)

fun named(name : String, point : Point3D) = NamedPoint3D(name = name, point = point)

fun canRoute(p0 : Point3D, p1 : Point3D) : Boolean  = !intersectsEarth(p0, p1)

fun <E> concat(set : Collection<E>, vararg more : E) : Set<E> {
    val mutable = set.toMutableSet()
    for (e in more) mutable.add(e)
    return mutable.toSet()
}

fun buildGraph(problem : Problem) : Graph {
    val start = named("Start", toCartesian(problem.route.startLat, problem.route.startLon))
    val end = named("End", toCartesian(problem.route.endLat, problem.route.endLon))

    val satellites = problem.satellites.map { sat ->
        named(sat.id, toCartesian(sat.lat, sat.lon, sat.height))
    }.toList()

    val startNode = Node(info = start, neighbors = satellites.filter {
        canRoute(start.point, it.point)
    }.map { it.name }.toSet())

    val allPoints = concat(satellites, start, end)

    val otherNodes = satellites.map { sat ->
        Node(info = sat, neighbors = allPoints.filter { other ->
            !sat.name.equals(other.name) && canRoute(sat.point, other.point)
        }.map { it.name }.toSet())
    }

    val endNode = Node(info = end, neighbors = satellites.filter {
        canRoute(it.point, end.point)
    }.map { it.name }.toSet())

    val nodeMap = concat(otherNodes, startNode, endNode).associateBy({ it.info.name }, { it })
    return Graph(nodeMap)
}

data class Routing (val node : String, var parent : String?, var distance : Int = Integer.MAX_VALUE)

fun solveGraph(start : String, end : String, graph : Graph) : List<String> {
    val routes : Map<String,Routing> = graph.nodes.map { Routing(node = it.key, parent = null) }.associateBy { it.node }
    val queue = LinkedList<String>()
    val route = routes[start]!!
    route.distance = 0
    queue.add(start)
    while (!queue.isEmpty()) {
        val current = queue.remove()
        val currentRoute = routes[current]!!
        val route = routes[current]!!
        val node = graph.nodes[current]!!
        for (neighbor in node.neighbors) {
            val neighborRoute = routes[neighbor]!!
            if (neighborRoute.distance == Integer.MAX_VALUE) {
                neighborRoute.distance = currentRoute.distance + 1
                neighborRoute.parent = current
                queue.add(neighbor)
            }
        }
    }
    val endRoute = routes[end]
    val nodes = LinkedList<String>()
    if (endRoute != null) {
        var current = endRoute
        while (current != null) {
            nodes.addFirst(current.node)
            val parent = current.parent
            if (parent != null) {
                current = routes[parent]
            } else {
                current = null
            }
        }

    }
    return nodes
}